require "optparse"

module TestRunner
  class CLI
    def self.run argv
      instance = new argv, $stdout
      instance.()
    end

    attr_reader :options

    def initialize argv, stdout
      @argv   = argv
      @stdout = stdout
    end

    def call
      @options = ArgvParser.(@argv)
      setup_config
      Runner.(filter_files) or exit 1
    end

    def setup_config
      Config.child_count = options.child_count
      Config.fail_fast = options.fail_fast?
      Config.logger = build_logger
      Config.reverse_backtraces = options.reverse_backtraces?
      Config.trim_backtraces = options.trim_backtraces?
    end

    def build_logger
      logger = Logger.new @stdout
      logger.level = options.log_level
      logger.progname = "test_runner"
      logger
    end

    def filter_files
      resolve_files.reject do |path|
        path.end_with? "init.rb"
      end
    end

    def resolve_files
      options.paths.flat_map do |path|
        if path.end_with? ".rb"
          [path]
        else
          Dir[File.join path, "**/*.rb"]
        end
      end
    end

    class Options
      def self.build
        if ENV["TEST_RUNNER_VERBOSE"]
          log_level = Logger::DEBUG
        else
          log_level = Logger::INFO
        end
        new log_level
      end

      attr_reader :child_count
      attr_reader :log_level
      attr_reader :paths

      def initialize log_level
        @paths = []
        @log_level = log_level
      end

      def add_path path
        paths << path
      end

      def child_count= number
        @child_count = number.to_i
      end

      def fail_fast
        @fail_fast = !@fail_fast
      end

      def fail_fast?
        @fail_fast
      end

      def full_backtrace
        @full_backtrace = !@full_backtrace
      end

      def trim_backtraces?
        return nil if @full_backtrace.nil?
        not @full_backtrace
      end

      def quiet
        @log_level += 1
      end

      def reverse_backtraces
        @reverse_backtraces = !@reverse_backtraces
      end

      def reverse_backtraces?
        @reverse_backtraces
      end

      def verbose
        @log_level -= 1
      end
    end

    class ArgvParser
      def self.call argv
        options = Options.build
        parser = ArgvParser.new argv, options
        parser.()
        options
      end

      def initialize argv, options
        @argv = argv
        @options = options
      end

      def call
        parse_options
        add_paths
      end

      def add_paths
        @argv << "tests" if @argv.empty?
        @argv.each do |path|
          @options.add_path path
        end
      end

      def parse_options
        OptionParser.new do |opts|
          opts.banner = <<-BANNER
Usage: #{program_name} [options] [PATH1] [PATH2] … [PATHn]

If no PATH(s) are specified, ./tests is assumed

          BANNER

          opts.on "-b", "--full-backtrace", "Do not filter assertion stack traces" do
            @options.full_backtrace
          end

          opts.on "-f", "--fail-fast", "When any test script fails, exit immediately" do
            @options.fail_fast
          end

          opts.on "-h", "--help", "Print this message and exit successfully" do
            puts opts
            exit 0
          end

          opts.on "-m", "--monkeypatch", "Turn on monkeypatching (equivalent to `require \"test_runner/script\"'" do
            require "test_runner/script"
          end

          opts.on "-n", "--num=COUNT", "Max number of sub processes to run concurrently" do |num|
            @options.child_count = num
          end

          opts.on "-q", "--quiet", "Reduces log verbosity" do
            @options.quiet
          end

          opts.on "-r", "--reverse-traces", "Reverse the order of stack traces" do
            @options.reverse_backtraces
          end

          opts.on "-V", "--version", "Print version and exit successfully" do
            spec = Gem.loaded_specs["test_runner"]
            version = if spec then spec.version else "(local)" end
            puts "Version: #{version}"
            exit 0
          end

          opts.on "-v", "--verbose", "Increases log verbosity" do
            @options.verbose
          end
        end.parse! @argv
      end

      def program_name
        File.basename $PROGRAM_NAME
      end
    end
  end
end
