module TestRunner
  class Runner
    def self.call paths
      instance = new paths
      instance.call
    end

    attr_reader :files
    attr_reader :pids

    def initialize files
      @files = files
    end

    def call
      Config.internal_logger.debug "test_runner found #{files.size} files: #{files * ", "}"
      return if files.empty?
      result = run_all

      log_msg = "finished executing files; success=#{result.inspect}"
      if result
        Config.logger.info log_msg
      else
        Config.logger.warn log_msg
      end

      result
    end

    def run_all
      set = ProcessSet.new
      Signal.trap "INT" do set.shutdown end
      set << files.shift until files.empty?
      set.finish
    end

    class ProcessSet
      def initialize
        @set     = []
        @passed  = true
      end

      def << file
        wait Config.child_count - 1
        @set.<< spawn_child file
      end

      def finish
        wait 0
        @passed
      end

      def wait max_count
        tick while @set.size > max_count
      end

      def tick
        loop do
          reads, _, _ = IO.select @set.map(&:fd), [], [], 1
          return reap reads if reads
        end
      end

      def spawn_child file
        process = Process.new file
        process.start
        process
      end

      def reap reads
        @set.delete_if do |process|
          next unless reads.include? process.fd
          Config.internal_logger.debug "Reaping #{process.file}:#{process.pid}"
          process.finish or @passed = false
          true
        end

        shutdown if failed? and Config.fail_fast
      end

      def shutdown
        @set.each do |process|
          ::Process.kill "TERM", process.pid
        end
        ::Process.waitall
        exit 1
      end

      def failed?
        not success?
      end

      def success?
        @passed
      end

      class Process
        attr_reader :fd
        attr_reader :file
        attr_reader :pid

        def initialize file
          @file = file
        end

        def start
          @fd, wr = IO.pipe

          @pid = fork do
            @fd.close
            begin
              load file
            rescue => error
              print_stacktrace error
              exit 1
            ensure
              wr.write "\x00"
            end
          end

          freeze
        end

        def print_stacktrace error
          locations = error.backtrace
          final_location = locations.shift
          locations = TestRunner::BacktraceFilter.(locations)

          lines = locations.map do |loc| "\tfrom #{loc}" end
          lines.unshift "#{final_location}: #{error.message} (#{error.class.name})"

          lines.reverse! if Config.reverse_backtraces
          Config.logger.error "Exception:\n#{lines * "\n"}"
        end

        def finish
          fd.read 1
          _, status = ::Process.wait2 pid
          status = status.exitstatus
          Config.internal_logger.debug "finished script #{@file}; status=#{status.inspect}"
          status == 0
        end
      end
    end
  end
end
